
/* Macros necessary to define user creep sets. */

// this controls body and weapon compatibility.
// A little explanation:
// w \ b | L | M | H  bodies can't be "ultra-", 
//    UL | + | - | -  ultra-light weapons are for light bodies only,
//     L | + | + | -  light weapons are for light or medium bodies,
//     M | + | + | +  medium weapons are for all bodies,
//     H | - | + | +  heavy weapons are for medium or heavy bodies,
//    UH | - | - | +  ultra-heavy weapons are for heavy bodies only.
WEIGHT = {
	ULTRALIGHT: 0,
	LIGHT: 1,
	MEDIUM: 2,
	HEAVY: 3,
	ULTRAHEAVY: 4
};

function creepWeightIsOK(bw, ww) {
	switch(ww) {
		case WEIGHT.ULTRALIGHT:
			return bw <= WEIGHT.LIGHT;
		case WEIGHT.LIGHT:
			return bw <= WEIGHT.MEDIUM;
		case WEIGHT.MEDIUM:
			return true;
		case WEIGHT.HEAVY:
			return bw >= WEIGHT.MEDIUM;
		case WEIGHT.ULTRAHEAVY:
			return bw >= WEIGHT.HEAVY;
	}
}

/* A useful auxiliary function for implementing
   sync-safe exponential growth. */

// approximately count floor[(a^n)/(b^{n-1})] synchronously,
// not relying on Math.pow() that may break sync.
function syncPower(a, b, n) {
	// some math here, not really worth understanding
	if (n == 0)
		return b;
	else if (n == 1)
		return a;
	else if (n % 2 == 1) {
		var i = a * syncPower(a, b, n - 1)
		return (i - (i % b)) / b;
	} else if (n % 2 == 0) {
		var p = syncPower(a, b, n / 2);
		var i = p * p;
		return (i - (i % b)) / b;
	}
}

/* 
   Call this function to actually send in the creeps.
   Designed to be repeatedly called from the user module,
   either directly or via creepsSpawnSmart() wrapper
   defined below.

   This function gives finer control over creeps. 
   creepsSpawnSmart(), on the other hand, does some
   work for you, generating creeps out of component lists.

   If `caption' is not empty string, sound will be played
   and console warning message will be displayed to the player.

 */
function creepsSpawn(
	player,        // enemy player to control the new creeps.
	position,      // where should the creeps come from.
	count,         // number of creeps to send in.
	bounty,        // money bonus for killing each creep.
	caption,       // player-visible description of creeps.
	body,          // creep body component
	propulsion,    // creep propulsion component
	weapon1,       // creep weapon component
	weapon2,       // creep second weapon, omit if not multi-turret
	weapon3        // creep third weapon, omit if not multi-turret
) {

	var name = "Droid for $" + bounty;

	var targets = enumStruct(0, HQ);
	// game over already?
	if (targets.length == 0)
		return;
	var target = targets[0];

	var droidType = DROID_WEAPON;
	if (propulsion == "CyborgLegs")
		droidType = DROID_CYBORG;

	// make sure at least $1 is received for each creep
	if (bounty <= 1)
		bounty = 1;

	// create creeps
	for (var i = 0; i < count; ++i) {
		if (typeof(weapon2) != "undefined") {
			if (typeof(weapon3) != "undefined") {
				addDroid(
					player, position.x, position.y, name,
					body, propulsion, 0, droidType,
					weapon1, weapon2, weapon3
				);
			} else {
				addDroid(
					player, position.x, position.y, name,
					body, propulsion, 0, droidType,
					weapon1, weapon2
				);
			}
		} else {
			addDroid(
				player, position.x, position.y, name,
				body, propulsion, 0, droidType,
				weapon1
			);
		}
	}

	// make creeps move
	enumDroid(player).forEach(function(droid) {
		orderDroidObj(droid, DORDER_ATTACK, target);
	});

	if (caption != "") {
		// notify the player
		playSound("pcv378.ogg", position.x, position.y, 0);
		console("Incoming " + caption);
	}
}

/* 

   You are to provide list of bodies, weapons and propulsions,
   and also count limitations, and this function would try to
   come up with some creep design.

   If multiple bodies (propulsions, weapons) are given, this function
   will prefer bodies (propulsions, weapons) from the end of the list,
   as long as minimum count of creeps can be achieved with given
   maximum value.

   This function would prefer a better body over a better propulsion,
   and also a better propulsion over a better weapon.

   If creepCount is positive, exactly creepCount creeps would be created.
   If creepCount is zero or negative, at least -creepCount creeps would be created.
   If creepCount is zero, at least one creep will be created anyway.

   The 'bodies' parameter is a list of {stat: ..., weight: ..., value: ...} objects.
   The 'propulsions' parameter is a list of {stat: ..., value: ...} objects.
   The 'bodies' parameter is a list of {stat: ..., weight: ..., value: ...} objects.

   Body weight is WEIGHT.LIGHT, WEIGHT.MEDIUM, or WEIGHT.HEAVY.
   Weapons can also have WEIGHT.ULTRALIGHT and WEIGHT.ULTRAHEAVY as their weights.

   Total value for a creep is sum of values for its body, propulsion and
   weapons. Total value for all creeps in a wave cannot exceed totalValue.

   Return value:
      * number of creeps created, 0 if no available design was found.

*/

function creepsSpawnSmart(
	player,        // enemy player to control the new creeps.
	position,      // where should the creeps come from.
	creepCount,    // number of creeps to send in.
	totalValue,    // maximum total value of creeps created.
	totalBounty,   // money bonus for killing all creeps of the whole wave.
	caption,       // player-visible description of creeps.
	bodies,        // list of bodies to choose from.
	propulsions,   // list of propulsions to choose from.
	weapons        // list of weapons to choose from.
) {

	var body = "";
	var propulsion = "";
	var weapon = "";
	var count = 0;
	var bestValue = 0;

	// handle negative creepCount
	var isMinCount = false;
	creepCount = Math.floor(creepCount);
	if (creepCount <= 0) {
		creepCount = -creepCount;
		if (creepCount == 0)
			creepCount = 1;
		isMinCount = true;
	}

	// choose creep design, as described above.
	for (var i = 0; i < bodies.length; ++i) {
		for (var j = 0; j < propulsions.length; ++j) {
			for (var k = 0; k < weapons.length; ++k) {
				if (!creepWeightIsOK(bodies[i].weight, weapons[k].weight))
					continue;
				var value = bodies[i].value + propulsions[j].value + weapons[k].value;
				if (creepCount * value < totalValue && value >= bestValue) {
					body = bodies[i].stat;
					propulsion = propulsions[j].stat;
					weapon = weapons[k].stat;
					bestValue = value;
					if (isMinCount)
						count = Math.floor(totalValue / value);
					else
						count = creepCount;
				}
			}
		}
	}

	// failed to find a suitable design
	if (count == 0)
		return 0;

	creepsSpawn(player, position, count, totalBounty / count, caption, body, propulsion, weapon);

	return count;

}
